@@ -1,7 +1,6 @@
 import re
+import sys
 
-from sphinx.ext.napoleon import Config
-from sphinx.ext.napoleon import docstring
 from docutils.core import publish_parts
 from docutils import nodes
 from hotdoc.core.comment_block import Comment
@@ -10,11 +9,16 @@ from docutils.writers.html4css1 import Writer as HtmlWriter
 from docutils.parsers.rst import roles, directives
 from xml.sax.saxutils import unescape
 
+from hotdoc_python_extension.napoleon import Config
+from hotdoc_python_extension.napoleon import docstring
+
 _google_typed_arg_regex = re.compile(r'\s*(.+?)\s*\(\s*(.+?)\s*\)')
 
 class MyGoogleDocString(docstring.GoogleDocstring):
     def __init__(self, *args, **kwargs):
         self.param_fields = []
+        self.attribute_fields = []
+        self.return_fields = []
         docstring.GoogleDocstring.__init__(self, *args, **kwargs)
 
     def _consume_field(self, parse_type=True, prefer_type=False):
@@ -54,11 +58,50 @@ class MyGoogleDocString(docstring.GoogleDocstring):
         self.param_fields.extend(self._consume_fields())
         return []
 
+    def _parse_attributes_section(self, section):
+        self.attribute_fields.extend(self._consume_fields())
+        return []
+
+    def _parse_returns_section(self, section):
+        self.return_fields.extend(self._consume_fields(prefer_type=True))
+        return []
+
 config = Config(napoleon_use_param=True, napoleon_use_rtype=True)
 
+def strip_doc(doc):
+    return '\n'.join([l.lstrip() for l in doc.split('\n')])
+
+# From http://stackoverflow.com/questions/2504411/proper-indentation-for-python-multiline-strings
+def trim(docstring):
+    if not docstring:
+        return ''
+    # Convert tabs to spaces (following the normal Python rules)
+    # and split into a list of lines:
+    lines = docstring.expandtabs().splitlines()
+    # Determine minimum indentation (first line doesn't count):
+    indent = sys.maxint
+    for line in lines[1:]:
+        stripped = line.lstrip()
+        if stripped:
+            indent = min(indent, len(line) - len(stripped))
+    # Remove indentation (first line is special):
+    trimmed = [lines[0].strip()]
+    if indent < sys.maxint:
+        for line in lines[1:]:
+            trimmed.append(line[indent:].rstrip())
+    # Strip off trailing and leading blank lines:
+    while trimmed and not trimmed[-1]:
+        trimmed.pop()
+    while trimmed and not trimmed[0]:
+        trimmed.pop(0)
+    # Return a single string:
+    return '\n'.join(trimmed)
+
 def google_doc_to_native(doc):
     if not doc:
-        return None
+        return (None, {})
+
+    doc = trim(doc)
 
     docstring = MyGoogleDocString(doc, config)
     comment = Comment(description=docstring.__unicode__(),
@@ -74,7 +117,29 @@ def google_doc_to_native(doc):
                 tags=tags)
         comment.params[field[0]] = param_comment
 
-    return comment
+    attr_comments = {}
+    for field in docstring.attribute_fields:
+        tags = {}
+        if field[1]:
+            tags['type'] = field[1]
+        prop_comment = Comment(name=field[0],
+                description='\n'.join(field[2]),
+                tags = tags)
+        attr_comments[field[0]] = prop_comment
+
+    return_comments = []
+    for field in docstring.return_fields:
+        tags = {}
+        if field[1]:
+            tags['type'] = field[1]
+        return_comment = Comment(
+                description='\n'.join(field[2]),
+                tags=tags)
+        return_comments.append(return_comment)
+
+    comment.tags['returns'] = return_comments
+
+    return comment, attr_comments
 
 class HotdocRestHtmlWriter(HtmlWriter):
     pass
@@ -113,7 +178,7 @@ def codeitem_directive(dirname, arguments, options, content,
         lineno, content_offset, block_set, state, state_machine):
     if not content:
         content = [u""]
-  
+
     m = _CALLABLE_RE.match(u"".join(arguments))
     m2 = _OTHER_RE.match(u"".join(arguments))
     if m:
@@ -188,23 +253,42 @@ directives.register_directive('envvar', codeitem_directive)
 
 def ref_role (name, raw_text, text, lineno, inliner,
         options=None, content=None):
-    doc_tool = inliner.document.settings.doc_tool
+    link_resolver = inliner.document.settings.link_resolver
+    cur_module = inliner.document.settings.cur_module
+
+    cur_module_components = cur_module.split('.')
+
     if options is None:
         options = {}
     if content is None:
         content = []
 
-    link = doc_tool.link_resolver.get_named_link(text)
+    text_components = text.split('.')
+
+    rel_distance = len(cur_module_components) - len(text_components) + 1
+
+    link = None
+
+    for i in range(rel_distance):
+        l = len(cur_module_components) - i
+        potential_name = '.'.join(cur_module_components[:l] + [text])
+        link = link_resolver.get_named_link(potential_name)
+        if link:
+            break
 
     if link is None:
-        node = nodes.Text(text)
+        link = link_resolver.get_named_link(text)
+
+    if link is None:
+        node = nodes.emphasis(text, text)
     else:
         node = nodes.reference(link.title, link.title, refuri=link.get_link(),
                 **options)
 
     return [node], []
 
-roles.register_local_role ('func', ref_role)
+roles.register_local_role('', ref_role)
+roles.register_local_role('func', ref_role)
 roles.register_local_role('mod', ref_role)
 roles.register_local_role('data', ref_role)
 roles.register_local_role('const', ref_role)
@@ -221,19 +305,16 @@ roles.register_local_role('ref', ref_role)
 
 
 class MyRestParser(object):
-    def __init__(self, extension, doc_tool=None):
-        self.doc_tool = doc_tool
+    def __init__(self, extension):
         self.extension = extension
         self.writer = HotdocRestHtmlWriter()
 
-    def translate(self, text, format_):
-        if format_ != 'html':
-            raise Exception("Unhandled format %s" % format_)
-        parts = publish_parts(text, writer=self.writer,
-                settings_overrides={'doc_tool': self.doc_tool})
-        return parts['fragment']
+    def translate(self, text, link_resolver, output_format, cur_module):
+        if output_format != 'html':
+            return text
 
-    def translate_comment(self, comment, format_):
-        print comment.filename
-        text = unescape(comment.description)
-        return self.translate(text, format_)
+        text = unescape(text)
+        parts = publish_parts(text, writer=self.writer,
+                settings_overrides={'link_resolver': link_resolver,
+                    'cur_module': cur_module})
+        return parts['fragment']
