@@ -1,17 +1,20 @@
-import os, ast, glob
+import os, glob
+import astroid as ast
+import pypandoc
 from hotdoc.core.base_extension import BaseExtension
+from hotdoc.core.file_includer import find_md_file
 from hotdoc.core.symbols import *
-from hotdoc.core.doc_tool import HotdocWizard
+from hotdoc.core.wizard import HotdocWizard
 from hotdoc.core.doc_tree import Page
 from hotdoc.utils.wizard import QuickStartWizard
 from hotdoc.core.comment_block import comment_from_tag
 
-from .python_doc_parser import google_doc_to_native, MyRestParser
+from .python_doc_parser import google_doc_to_native
 from .python_html_formatter import PythonHtmlFormatter
 
 class PythonScanner(object):
-    def __init__(self, doc_tool, extension, sources):
-        self.doc_tool = doc_tool
+    def __init__(self, doc_repo, extension, sources):
+        self.doc_repo = doc_repo
 
         self.class_nesting = 0
 
@@ -23,13 +26,19 @@ class PythonScanner(object):
                 }
 
         self.__extension = extension
+        self.mod_comments = {}
 
         for source in sources:
+            relpath = os.path.relpath(source, self.__extension.package_root)
+            # FIXME: ahem
+            modname = os.path.splitext(relpath)[0].replace('/', '.')
             self.__current_filename = source
-            with open (source, 'r') as f:
-                code = f.read()
-            tree = ast.parse(code)
-            modname = os.path.basename (os.path.splitext(source)[0])
+            builder = ast.builder.AstroidBuilder()
+            tree = builder.file_build(source)
+            modcomment, attribute_comments = google_doc_to_native(tree.doc)
+            if modcomment:
+                self.mod_comments[modname] = modcomment
+
             self.__parse_module (tree.body, modname)
 
     def __create_fundamentals(self):
@@ -61,6 +70,10 @@ class PythonScanner(object):
                 Link('https://docs.python.org/2/tutorial/datastructures.html#dictionaries',
                         'dict', None)
 
+        callable_link = \
+                Link('https://docs.python.org/2/library/functions.html#callable',
+                        'callable', None)
+
         fundamentals = {
                 "none": none_link,
                 "boolean": boolean_link,
@@ -76,6 +89,7 @@ class PythonScanner(object):
                 "False": false_link,
                 "false": false_link,
                 "dict": dict_link,
+                "callable": callable_link,
                 "dictionary": dict_link,
         }
 
@@ -90,24 +104,63 @@ class PythonScanner(object):
     def __parse_class (self, klass, parent_name):
         self.class_nesting += 1
         klass_name = '.'.join((parent_name, klass.name))
-        docstring = ast.get_docstring(klass)
-        comment = google_doc_to_native(docstring)
+        comment, attr_comments = google_doc_to_native(klass.doc)
 
         if comment:
             comment.filename = self.__current_filename
 
+        for method in sorted(klass.mymethods(), key=lambda x: x.name):
+            self.__parse_function(method, klass_name, is_method=True)
+
+        for method in klass.methods():
+            if method.name == '__init__' and method.parent.parent.name != \
+                    '__builtin__':
+                self.__parse_function(method, klass_name, is_method=True,
+                        is_ctor_for=klass_name)
+
+        for attr_name, attr in sorted(klass.instance_attrs.items()):
+            attr_comment = attr_comments.get(attr_name)
+            self.__parse_attribute(klass_name, attr_comment, attr_name, attr)
+
         class_symbol = self.__extension.get_or_create_symbol(ClassSymbol,
                 comment=comment,
                 filename=self.__current_filename,
                 display_name=klass_name)
 
-        for node in klass.body:
-            f = self.__node_parsers.get (type(node))
-            if f:
-                f(node, klass_name)
-
         self.class_nesting -= 1
 
+    def __type_tokens_from_comment(self, comment):
+        if comment is None:
+            return []
+
+        try:
+            pytype = comment.tags.pop('type')
+        except KeyError:
+            return []
+
+        try:
+            link = self.fundamentals[pytype]
+        except KeyError:
+            link = Link(None, pytype, pytype)
+
+        return [link]
+
+    def __parse_attribute(self, parent_name, attr_comment, attr_name, attr):
+        if attr_name.startswith('__'):
+            return
+
+        attr_name = '.'.join((parent_name, attr_name))
+
+        type_tokens = self.__type_tokens_from_comment(attr_comment)
+
+        type_ = QualifiedSymbol(type_tokens=type_tokens)
+
+        self.__extension.get_or_create_symbol(PropertySymbol,
+                comment=attr_comment,
+                filename=self.__current_filename,
+                display_name=attr_name,
+                prop_type=type_)
+
     def __params_doc_to_dict (self, params_doc):
         dict_ = {}
         for param in params_doc:
@@ -115,25 +168,24 @@ class PythonScanner(object):
 
         return dict_
 
-    def __parse_function (self, function, parent_name):
+    def __parse_function (self, function, parent_name,
+            is_method=False, is_ctor_for=None):
+        if function.name.startswith('__'):
+            return
+
+        if not is_method and function.name.startswith('_'):
+            return
+
         func_name = '.'.join ((parent_name, function.name))
 
-        docstring = ast.get_docstring (function)
-        if docstring:
-            comment = google_doc_to_native(docstring)
+        if function.doc:
+            comment, attr_comments = google_doc_to_native(function.doc)
             comment.filename = self.__current_filename
         else:
             comment = None
 
         parameters = self.__parse_parameters(function.args, comment)
-
-        if comment:
-            return_tag = comment.tags.get('returns')
-            return_comment = comment_from_tag(return_tag)
-            retval = ReturnValueSymbol (type_tokens=[],
-                    comment=return_comment)
-        else:
-            retval = None
+        retval = self.__parse_return_value(comment)
 
         is_method = self.class_nesting > 0
 
@@ -145,29 +197,38 @@ class PythonScanner(object):
                 return_value=retval,
                 comment=comment,
                 is_method = is_method,
+                is_ctor_for = is_ctor_for,
                 filename=self.__current_filename,
                 display_name=func_name)
 
+    def __parse_return_value(self, comment):
+        if not comment:
+            return [None]
+
+        try:
+            ret_comments = comment.tags.pop('returns')
+            return_value = []
+            for ret_comment in ret_comments:
+                type_tokens = self.__type_tokens_from_comment(ret_comment)
+                return_value.append(ReturnItemSymbol(type_tokens=type_tokens,
+                        comment=ret_comment))
+            return return_value or [None]
+        except KeyError:
+            return [None]
+
     def __parse_parameters(self, args, comment):
         parameters = []
+
         if comment:
             param_comments = comment.params
         else:
             param_comments = {}
 
-        for arg in args.args:
-            param_comment = param_comments.get (arg.id)
-            type_tokens = []
-            if param_comment:
-                type_ = param_comment.tags.pop('type', None)
-                if type_:
-                    link = self.fundamentals.get(type_)
-                    if link:
-                        type_tokens = [link]
-                    else:
-                        type_tokens = [type_]
+        for arg in args.args or []:
+            param_comment = param_comments.get (arg.name)
+            type_tokens = self.__type_tokens_from_comment(param_comment)
 
-            param = ParameterSymbol (argname=arg.id,
+            param = ParameterSymbol (argname=arg.name,
                     type_tokens=type_tokens,
                     comment=param_comment)
             parameters.append (param)
@@ -204,6 +265,9 @@ These wildcards will be evaluated each time hotdoc is run.
 """
 
 def validate_filters(wizard, thing):
+    if thing is None:
+        return True
+
     if not QuickStartWizard.validate_globs_list(wizard, thing):
         return False
 
@@ -226,8 +290,8 @@ def resolve_patterns(source_patterns, conf_path_resolver):
     return source_files
 
 def source_files_from_config(config, conf_path_resolver):
-    sources = resolve_patterns(config.get('python_sources', []), conf_path_resolver)
-    filters = resolve_patterns(config.get('python_source_filters', []),
+    sources = resolve_patterns(config.get('python_sources') or [], conf_path_resolver)
+    filters = resolve_patterns(config.get('python_source_filters') or [],
             conf_path_resolver)
     sources = [item for item in sources if item not in filters]
     return [os.path.abspath(source) for source in sources]
@@ -235,46 +299,41 @@ def source_files_from_config(config, conf_path_resolver):
 class PythonExtension(BaseExtension):
     EXTENSION_NAME = 'python-extension'
 
-    def __init__(self, doc_tool, config):
-        BaseExtension.__init__(self, doc_tool, config)
-        self._doc_parser = MyRestParser(self, doc_tool)
-        self.sources = source_files_from_config(config, doc_tool)
-        self.package_root = os.path.commonprefix(self.sources)
+    def __init__(self, doc_repo, config):
+        BaseExtension.__init__(self, doc_repo, config)
+        self.sources = source_files_from_config(config, doc_repo)
+
+        self.package_root = config.get('python_package_root')
+        if not self.package_root:
+            self.package_root = os.path.commonprefix(self.sources)
+        self.package_root = os.path.abspath(os.path.join(self.package_root,
+            '..'))
+
         self.python_index = config.get('python_index')
-        doc_tool.doc_tree.page_parser.register_well_known_name('python-api',
+        doc_repo.doc_tree.page_parser.register_well_known_name('python-api',
                 self.python_index_handler)
-        self._formatters['html'] = PythonHtmlFormatter(self.doc_tool, self)
+        self.formatters['html'] = PythonHtmlFormatter(
+            self, doc_repo.doc_database)
 
     def setup(self):
-        if not self.stale_source_files:
+        stale, unlisted = self.get_stale_files(self.sources)
+        if not stale:
             return
 
-        self.scanner = PythonScanner (self.doc_tool, self,
-                self.stale_source_files)
+        self.stale = stale
+
+        self.scanner = PythonScanner (self.doc_repo, self,
+                stale)
 
         if not self.python_index:
-            index_path = self.create_naive_index()
-            old_prefix = self.doc_tool.doc_tree.page_parser.prefix
-            self.doc_tool.doc_tree.page_parser.prefix = os.path.dirname(index_path)
-            new_page = self.doc_tool.doc_tree.build_tree(index_path,
-                    'python-extension')
-            self.doc_tool.doc_tree.page_parser.prefix = old_prefix
-            self.doc_tool.doc_tree.pages['python-extension-index.markdown'] = new_page
-
-    def get_source_files(self):
-        return self.sources
+            self.update_naive_index()
 
     def python_index_handler (self, doc_tree):
         if not self.python_index:
-            new_page = Page('python-extension-index')
-            doc_tree.pages['python-extension-index.markdown'] = new_page
-            doc_tree.page_parser.parse_contents(new_page, '')
-            return 'python-extension-index.markdown'
+            return self.create_naive_index(self.sources)
 
-        index_path = os.path.join(doc_tree.prefix, self.python_index)
-        index_path = self.doc_tool.resolve_config_path(index_path)
-        new_page = doc_tree.build_tree(index_path, 'python-extension')
-        return index_path, ''
+        index_path = find_md_file(self.python_index, self.doc_repo.include_paths)
+        return index_path, '', 'python-extension'
 
     @staticmethod
     def add_arguments (parser):
@@ -290,10 +349,33 @@ class PythonExtension(BaseExtension):
                 extra_prompt=PYTHON_FILTERS_PROMPT,
                 validate_function=validate_filters,
                 finalize_function=HotdocWizard.finalize_paths)
+        group.add_argument ("--python-package-root", action="store", nargs="+",
+                dest="python_package_root", help="Path to the root of the"
+                " documented package / application",
+                validate_function=HotdocWizard.validate_folder,
+                finalize_function=HotdocWizard.finalize_path)
         group.add_argument ("--python-index", action="store",
                 dest="python_index",
                 help="Path to the python root markdown file",
                 finalize_function=HotdocWizard.finalize_path)
 
+    def _get_naive_link_title(self, source_file):
+        relpath = os.path.relpath(source_file, self.package_root)
+        modname = os.path.splitext(relpath)[0].replace('/', '.')
+        return modname
+
+    def _get_naive_page_description(self, link_title):
+        modcomment = self.scanner.mod_comments.get(link_title)
+        if modcomment is None:
+            return ''
+
+        if modcomment.description:
+            out = '## %s\n\n' % link_title
+            out += pypandoc.convert(modcomment.description, to='md',
+                    format='rst')
+            return out
+
+        return BaseExtension._get_naive_page_description(self, link_title)
+
 def get_extension_classes():
     return [PythonExtension]
